(******************************************************************************

______________________________________________________________________________

PDB2TXT                                                   (C) 2009, 2010 Pepak
http://www.pepak.net/e-books/formaty-pdb/                 http://www.pepak.net
______________________________________________________________________________


Copyright (c) 2010, Pepak (http://www.pepak.net)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Pepak nor the
      names of his contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL PEPAK BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

******************************************************************************)

unit uPdb2txt;

interface

uses
  SysUtils, Classes,
  DocFile, PalmDatabase,
  uCompatibility, uConsoleApp, uStringUtils;

type
  TEncoding = (encANSI, encUTF8, encUTF16);
  
  TPdb2Txt = class(TConsoleApp)
    protected
      function AppTitle: string; override;
      function AppVersion: string; override;
      function DoExecute: integer; override;
      procedure ShowSyntax(const Error: string = ''); override;
      procedure ParamInitialize; override;
    private
      fRecursive: boolean;
      fBOM: boolean;
      fEncoding: TEncoding;
    protected
    public
      constructor Create; override;
      destructor Destroy; override;
      function ConvertFile(const FileName: string): integer; virtual;
      function ConvertFiles(const WildCard: string): integer; virtual;
      property Recursive: boolean read fRecursive write fRecursive;
      property BOM: boolean read fBOM write fBOM;
      property Encoding: TEncoding read fEncoding write fEncoding;
    end;

const
  EXITCODE_OK = 0;
  EXITCODE_NO_DATA = 1;
  EXITCODE_CONVERSION_FAILED = 2;
  EXITCODE_SYNTAX = 254;

implementation

{ TPdb2Txt }

constructor TPdb2Txt.Create;
begin
  inherited;
end;

destructor TPdb2Txt.Destroy;
begin
  inherited;
end;

function TPdb2Txt.AppTitle: string;
begin
  Result := 'PDB2TXT';
end;

function TPdb2Txt.AppVersion: string;
begin
  Result := {$INCLUDE 'Pdb2Txt.version'} ;
end;

procedure TPdb2Txt.ShowSyntax(const Error: string);
begin
  inherited;
  WriteColored(ccWhite, '[<arg>] [<arg>] ... <wildcard> [<wildcard>] ...'); Writeln;
  Writeln;
  WriteColored(ccWhite, ' -h, -?'); Writeln(' ... Show this help screen.');
  WriteColored(ccWhite, ' -r'); Writeln(' ....... Scan directories recursively.');
  WriteColored(ccWhite, ' -r-'); Writeln(' ...... Don''t scan directories recursively (default).');
  WriteColored(ccWhite, ' -ansi'); Writeln(' .... Use ANSI encoding for output (default).');
  WriteColored(ccWhite, ' -utf8'); Writeln(' .... Use UTF-8 encoding for output.');
  WriteColored(ccWhite, ' -utf16'); Writeln(' ... Use UTF-16 encoding ("Unicode" in Notepad) for output.');
  WriteColored(ccWhite, ' -bom'); Writeln(' ..... Write BOM at the beginning of files (default).');
  WriteColored(ccWhite, ' -bom-'); Writeln(' .... Don''t write BOM at the beginning of files.');
  Writeln;
end;

procedure TPdb2Txt.ParamInitialize;
begin
  inherited;
  Recursive := False;
  BOM := True;
  Encoding := encANSI;
end;

function TPdb2Txt.DoExecute: integer;

  procedure SyntaxError(const Msg: string); overload;
    begin
      ShowSyntax(Msg);
      Abort;
    end;
  procedure SyntaxError(const Msg: string; const Params: array of const); overload;
    begin
      SyntaxError(Format(Msg, Params));
    end;

var Param: string;
    Code: integer;
    MaskFound: boolean;
begin
  Result := EXITCODE_NO_DATA;
  MaskFound := False;
  ParamInitialize;
  try
    while ParamGetNext(Param) do
      if Param[1] = '-' then
        if (Param = '-h') or (Param = '-?') then
          SyntaxError('')
        else if Param = '-r' then
          Recursive := True
        else if Param = '-r-' then
          Recursive := False
        else if Param = '-bom' then
          BOM := True
        else if Param = '-bom-' then
          BOM := False
        else if Param = '-ansi' then
          Encoding := encANSI
        else if Param = '-utf8' then
          Encoding := encUTF8
        else if Param = '-utf16' then
          Encoding := encUTF16
        else
          SyntaxError('Unknown argument "%s".', [Param])
      else
        begin
        MaskFound := True;
        Code := ConvertFiles(Param);
        if Code <> EXITCODE_NO_DATA then
          if (Code <> EXITCODE_OK) or (Result = EXITCODE_NO_DATA) then
            Result := Code;
        end;
    if not MaskFound then
      Result := ConvertFiles('*.pdb');
    if Result = EXITCODE_NO_DATA then
      ShowSyntax('No files found.');
  except
    on EAbort do
      Result := EXITCODE_SYNTAX;
    end;
end;

function TPdb2Txt.ConvertFiles(const WildCard: string): integer;
var Path, Mask: string;
    SR: TSearchRec;
    Code: integer;
begin
  WriteColored(ccLightCyan, WildCard);
  Writeln;
  Result := EXITCODE_NO_DATA;
  Path := ExtractFilePath(WildCard);
  Mask := ExtractFileName(WildCard);
  if FindFirst(WildCard, faAnyFile and (not faDirectory), SR) = 0 then
    try
      repeat
        if LongBool(SR.Attr and faDirectory) then
          begin
          // Nothing to do. I should never get here anyway.
          end
        else
          begin
          Code := ConvertFile(Path + SR.Name);
          if Code <> EXITCODE_NO_DATA then
            if (Code <> EXITCODE_OK) or (Result = EXITCODE_NO_DATA) then
              Result := Code;
          end;
      until FindNext(SR) <> 0;
    finally
      FindClose(SR);
      end;
  if Recursive then
    if FindFirst(Path + '*.*', faDirectory, SR) = 0 then
      try
        repeat
          if LongBool(SR.Attr and faDirectory) then
            begin
            if (SR.Name <> '.') and (SR.Name <> '..') then
              begin
              Code := ConvertFiles(Path + SR.Name + '\' + Mask);
              if Code <> EXITCODE_NO_DATA then
                if (Code <> EXITCODE_OK) or (Result = EXITCODE_NO_DATA) then
                  Result := Code;
              end;
            end;
        until FindNext(SR) <> 0;
      finally
        FindClose(SR);
        end;
end;

function TPdb2Txt.ConvertFile(const FileName: string): integer;
var PDB: TDocFile;
    FS: TFileStream;
    FN: string;
    Text, BomString: AnsiString;
    Utf8Text: Utf8String;
    Utf16Text: WideString;
begin
  Result := EXITCODE_OK;
  Write(Format('%-70s  ', [FileName]));
  try
    PDB := TDocFile.Create(FileName);
    try
      Text := PDB.GetDocument;
      FN := FileName + '.txt';
      try
        if FileExists(FN) then
          DeleteFile(FN);
        FS := TFileStream.Create(FN, fmCreate or fmShareDenyWrite);
        try
          case Encoding of
            encAnsi:
              FS.WriteBuffer(Text[1], Length(Text));
            encUTF8:
              begin
              if BOM then
                begin
                BomString := #$EF#$BB#$BF;
                FS.WriteBuffer(BomString[1], Length(BomString));
                end;
              Utf8Text := WideToUtf8(AnsiToWide(Text));
              FS.WriteBuffer(Utf8Text[1], Length(Utf8Text));
              end;
            encUTF16:
              begin
              if BOM then
                begin
                BomString := #$FF#$FE;
                FS.WriteBuffer(BomString[1], Length(BomString));
                end;
              Utf16Text := WideString(Text);
              FS.WriteBuffer(Utf16Text[1], Length(Utf16Text)*Sizeof(Utf16Text[1]));
              end;
            end;
          WriteColored(ccWhite, 'OK');
          Writeln;
        finally
          FS.Free;
          end;
      except
        if FileExists(FN) then
          DeleteFile(FN);
        Raise;
        end;
    finally
      PDB.Free;
      end;
  except
    WriteColored(ccLightRed, 'FAILED');
    Writeln;
    Result := EXITCODE_CONVERSION_FAILED;
    end;
end;

end.
